SuperStrict
Framework BRL.GLMax2D
Import BRL.D3D7Max2D
Import BRL.D3D9Max2D
Import BRL.EventQueue
Import BRL.LinkedList
Import BRL.FileSystem
Import BRL.PngLoader
Import BRL.MaxLUA
Import BRL.Timer
Import PUB.FreeType
Import BRL.Font
Import BRL.FreeTypeFont



Const sx:Int = 800, sy:Int = 600
Graphics sx, sy
SetBlend ALPHABLEND
Local font:TImageFont = LoadImageFont("font.ttf", 16)
If Not font Then DebugLog "font = Null"
SetImageFont font


Const DELAY_TIME:Int = 5000
Const NO_KEYS:Int = 3

Const MSG_HEIGHT:Int = 40
Const MSG_PADDING:Int = 80


Const MSG_MAX_LEN:Int = 50
Const MSGNEW_HEIGHT:Int = 40

Const DISPLAY_TIME:Int = 200

Global msgBg:TImage = LoadImage("msgBg.png")
Global senderBg:TImage = LoadImage("senderBg.png")
Global receiverBg:TImage = LoadImage("receiverBg.png")

Global currentState:TGameState

Global online:TList = New TList
Global caught_msg:TList = New TList
Global messages:TList = New TList
Global playerKeys:Int[NO_KEYS]

Global keyImg:TImage = LoadImage("key.png")
Global lockImg:TImage = LoadImage("lock.png")
Global keyR:Int[NO_KEYS], keyG:Int[NO_KEYS], keyB:Int[NO_KEYS]
Function setkeycol(r:Int, g:Int, b:Int)

	Global index:Int
	keyR[index] = r
	keyG[index] = g
	keyB[index] = b
	index:+1

End Function
setkeycol 0, 0, 255
setkeycol 0, 255, 0
setkeycol 255, 215, 0

Global savefile$ = AppFile+".save"

Global events:TList = New TList
LuaRegisterObject New TLUAAPI, "API"
Global luaClass:TLuaClass = TLuaClass.Create(LoadString("story.lua"))
Global luaObject:TLuaObject = TLuaObject.Create(luaClass, Null)

Function GetComputer:TComputer(name$)

	For Local c:TComputer = EachIn TComputer.list

		If c.name = name Then Return c

	Next

End Function
Type TLuaEvent

	Field func$, run:Int, gameStateDependant:Int

End Type
Function RunEvents()

	Local time:Int = MilliSecs()
	For Local e:TLuaEvent = EachIn events

		If e.run <= time Then

			luaObject.Invoke(e.func, Null)
			events.Remove e

		EndIf

	Next

End Function
			
Global gsSend:Int, gsSelectSender:Int, gsCancel:Int
Global gsKeys:Int[NO_KEYS]

Global lockNotification:Int = False

Type TLUAAPI

	Field currentState:Int
	Method ChangeState(num:Int)

		currentState = num
		.ChangeState TGameState.array[num]

	End Method
	Method Notify(s$)

		SendNotification s

	End Method
	Method ActiveNotification:Int()

		Return .activeNotification

	End Method
	Method SendMsg:Int(sender$, receiver$, content$, key:Int)

		Local s:TComputer = GetComputer(sender)
		Local r:TComputer = GetComputer(receiver)
		If Not s.online Or Not r.online Then

			DebugLog "sender: "+sender+" ("+s.online+")"	
			DebugLog "receiver: "+receiver+" ("+r.online+")"	
			RuntimeError "tried to send message from or to offline computer"
		EndIf

		s.SendMsg r, content, key

	End Method

	Method MsgAvailable:Int(comp$)

		Return Not GetComputer(comp).new_msg.IsEmpty()

	End Method
	Method MsgSender$(comp$)

		Return TMessage(GetComputer(comp).new_msg.First()).sender.name

	End Method
	Method MsgContent$(comp$)

		Return TMessage(GetComputer(comp).new_msg.First()).content

	End Method
	Method MsgKey:Int(comp$)

		Return TMessage(GetComputer(comp).new_msg.First()).key

	End Method
	Method MsgNext:Int(comp$)

		Local c:TComputer = GetComputer(comp)
		If c.new_msg.IsEmpty() Then Return False
		Local f:Object = c.new_msg.First()
		c.new_msg.Remove f
		Return Not c.new_msg.IsEmpty()

	End Method

	Method MsgPop:Int(comp$)

		Local c:TComputer = GetComputer(comp)
		If c.new_msg.IsEmpty() Then Return False
		Local f:Object = c.new_msg.First()
		c.new_msg.Remove f
		Return True

	End Method
	Method SetTimeout(func$, ms:Int, gameStateDependant:Int)

		Local e:TLuaEvent = New TLuaEvent
		e.func = func
		e.run = MilliSecs()+ms
		e.gameStateDependant = gameStateDependant
		events.AddLast e

	End Method

	Method CreateComputer(name$, r:Int, g:Int, b:Int)

		Local c:TComputer = New TComputer
		c.Set name, r, g, b
		
	End Method
	Method CreateGameState:Int()

		Local s:TGameState = New TGameState
		s.permSend = gsSend
		s.permSelectSender = gsSelectSender
		s.permCancel = gsCancel
		s.keys = gsKeys[..]
		Return s.num

	End Method
	Method GameStateOnline(num:Int, comp$)

		TGameState.array[num].onlineComputers.AddLast GetComputer(comp)

	End Method
	Method AddPermission()

		If Not gsSend Then

			gsSend = True

		ElseIf Not gsCancel Then

			gsCancel = True

		Else
			gsSelectSender = True

		EndIf

	End Method
	Method AddKey()

		For Local i:Int = 0 Until NO_KEYS

			If Not gsKeys[i] Then

				gsKeys[i] = True
				Exit

			EndIf

		Next

	End Method

	Method LockNotification()

		.lockNotification = True

	End Method

	Method GoOnline(comp$)

		GetComputer(comp).GoOnline

	End Method
	Method GoOffline(comp$)

		GetComputer(comp).GoOffline

	End Method

End Type


Global invokeIncr:Int

Type TGameState

	Global array:TGameState[]
	Method New()

		num = array.length
		array :+ [Self]

	End Method

	Field num:Int
	Field active:Int = False
	Field onlineComputers:TList = New TList
	Field permSend:Int, permSelectSender:Int, permCancel:Int
	Field keys:Int[]


	Method Init()

		If currentState Then

			currentState.Terminate

		EndIf
		currentState = Self
		active = True
		For Local c:TComputer = EachIn onlineComputers

			c.GoOnline

		Next
		.permSend = permSend
		.permSelectSender = permSelectSender
		.permCancel = permCancel
		playerKeys = keys
		messages.Clear
		luaObject.Invoke "init_"+num, Null
		
		selectedSender = playerComputer
		selectedReceiver = Null
		selectedKey = 0

	End Method

	Method Terminate()

		For Local c:TComputer = EachIn online

			c.GoOffline

		Next
		For Local e:TLuaEvent = EachIn events
		
			If e.gameStateDependant Then

				events.Remove e

			EndIf

		Next
		active = False

	End Method

	Method Update()

		invokeIncr :+ 1
		If invokeIncr > 6 Then

			invokeIncr = 0
			luaObject.Invoke "gameState_"+num, Null

		EndIf
		RunEvents

		For Local m:TMessage = EachIn caught_msg

			m.Update

		Next
		If Not active Then

			Return

		EndIf

	End Method

	Function Load()

		
		If FileType(savefile) = 0 Then

			array[0].Init

		Else

			Local f:TStream = ReadFile(savefile)
			array[f.ReadInt()].Init()
			f.Close

		EndIf

	End Function
	Function Save:Int()

		Local f:TStream = WriteFile(savefile)
		f.WriteInt currentState.num
		f.Close

	End Function

End Type

Function ChangeState(s:TGameState)

	s.Init

End Function

Type TComputer

	Global list:TList = New TList
	Method New()

		list.AddLast Self

	End Method

	Field name$
	Field new_msg:TList = New TList
	Field online:Int
	Field r:Int, g:Int, b:Int

	Method Set(name$, r:Int, g:Int, b:Int)

		Self.name = name
		Self.r = r
		Self.g = g
		Self.b = b

	End Method

	Method GoOnline()

		If Not online Then

			online = True
			.online.AddLast Self

		EndIf
		new_msg.Clear

	End Method

	Method GoOffline()

		If online Then

			online = False
			.online.Remove Self

		EndIf
		new_msg.Clear

	End Method

	Method SendMsg(receiver:TComputer, content$, key:Int = 0, sender:TComputer = Null)

		If Not online Then RuntimeError "Tried to send message of offline computer"
		If Not sender Then sender = Self
		Local msg:TMessage = CreateMessage(sender, receiver, key, content)

	End Method

End Type

Function CreateMessage:TMessage(sender:TComputer, receiver:TComputer, key:Int, content$)

	Local result:TMessage = New TMessage
	result.sender = sender
	result.receiver = receiver
	result.key = key
	result.content = content
	caught_msg.AddFirst result
	messages.AddFirst result
	result.created = MilliSecs()
	If receiver = playerComputer Then

		result.display = True

	Else

		result.caught = True

	EndIf
	Return result

End Function

Type TMessage

	Field sender:TComputer, receiver:TComputer
	Field key:Int, content$
	Field created:Int
	Field caught:Int, cancelled:Int

	Field display:Int, space#
	

	Method Update()

		If caught Then
			Local time:Int = MilliSecs()
			If Not display Then

				Local s# = (1.0/DISPLAY_TIME)*(time-created)
				If s >= 1 Then

					display = True

				Else

					space = MSG_HEIGHT*(0.5-0.5*Cos(180*s))

				EndIf

			ElseIf time-DELAY_TIME >= created Then
	
				caught = False
				caught_msg.Remove Self
				receiver.new_msg.AddLast Self

			EndIf

		EndIf

	End Method

	Method Cancel()

		If caught Then

			caught = False
			caught_msg.Remove Self
			cancelled = True

		EndIf
		display = True

	End Method

	Method Draw(y#)

		Local s#

		If caught Then

			s = (1.0/DELAY_TIME)*(MilliSecs()-created)
			If s > 0.5 Then
			
				SetColor 200, 200+55*2*(1-s), 255
				
			Else
			
				SetColor 200, 255, 200+55*s*2
				
			EndIf

		ElseIf Self.cancelled Then

			SetColor 255, 100, 100

		Else

			SetColor 220, 220, 220

		EndIf
		DrawImage msgBg, 0, y

		If caught Then

			SetColor 255, 255, 255
			Local x# = MSG_PADDING + (sx-2*MSG_PADDING)*s
			DrawLine x, y, x, y+MSG_HEIGHT

		EndIf

		If Not key Or playerKeys[key-1] Then
		
			SetColor 0, 0, 0
			DrawText content, (sx-TextWidth(content))/2, y+(MSG_HEIGHT-TextHeight(content))/2
			If key Then

				SetColor keyR[key-1], keyG[key-1], keyB[key-1]
				DrawImage keyImg, MSG_PADDING, y

			EndIf

		Else

			SetColor 255, 255, 255
			DrawImage lockImg, (sx-ImageWidth(lockImg))/2, y

		EndIf

		SetColor sender.r, sender.g, sender.b
		DrawImage senderBg, 0, y
		SetColor 255, 255, 255
		DrawText sender.name, (MSG_PADDING-TextWidth(sender.name))/2, y+(MSG_HEIGHT-TextHeight(sender.name))/2

		SetColor receiver.r, receiver.g, receiver.b
		DrawImage receiverBg, sx-ImageWidth(receiverBg), y
		SetColor 255, 255, 255
		DrawText receiver.name, sx-(MSG_PADDING+TextWidth(receiver.name))/2, y+(MSG_HEIGHT-TextHeight(receiver.name))/2

	End Method

	Method Parse$(key$)

		Local pos:Int = content.Find(key+"=")
		If pos=-1 Then Return ""
		pos :+ key.length+1
		Local e:Int = content.Find(";")
		If e=-1 Then

			e = content.length

		EndIf
		Return content[pos..e]

	End Method

End Type

Function SendNotification(n$)

	notification = Null
	Local pos:Int = 0
	Repeat

		Local p:Int = n.Find("~n", pos)
		If p<>-1 Then

			notification :+ [n[pos..p].Trim()]
			pos = p+1

		Else

			Exit

		EndIf

	Forever
	notification :+ [n[pos..], "", "Click right to close"]
	activeNotification = True
	notiFadeStart = MilliSecs()

End Function

Type TPlayerComputer Extends TComputer

	Method New()

		Set "Me", 150, 150, 150

	End Method

End Type

Global playerComputer:TPlayerComputer = New TPlayerComputer
LuaRegisterObject playerComputer, "player"

Function UpdateEverything()

	currentState.Update


End Function

Global notification$[]
Global activeNotification:Int
Global notiFadeStart:Int
Const NOTIFICATION_FADE:Int = 500

Global selectedSender:TComputer = playerComputer, selectedReceiver:TComputer, inputContent$, blink:Int
Global selectedKey:Int
Global selectSender:Int, selectReceiver:Int, selectKey:Int

Function DrawEverything()

	Local y# = 0
	For Local msg:TMessage = EachIn messages

		If y >= sy-MSGNEW_HEIGHT Then

			If Not msg.caught Then messages.Remove msg
			Continue

		EndIf
	
		If msg.display Then

			msg.Draw y
			y :+ MSG_HEIGHT
			
		Else

			y :+ msg.space

		EndIf
	Next

	If notification Then

		Local time:Int = MilliSecs()
		Local a# = (1.0/NOTIFICATION_FADE)*(time-notiFadeStart)
		If Not activeNotification Then a = 1-a
		SetAlpha a
		Local height# = 20*notification.length
		Local width#
		For Local l$ = EachIn notification

			width = Max(width, TextWidth(l))

		Next
		Local x# = (sx-width)*0.5
		Local y# = (sy-height)*0.5
		SetColor 100, 100, 100
		DrawRect x-3, y-3, width+6, height+6
		SetColor 255, 255, 255
		For Local l$ = EachIn notification

			DrawText l, x, y
			y :+ 20

		Next
		If a < 0 Then

			notification = Null
			SetAlpha 1

		ElseIf a < 1 Then

			SetAlpha 1

		EndIf

	EndIf


	If Not permSend Then Return
	SetColor 100, 100, 100
	DrawRect 0, sy-MSGNEW_HEIGHT, sx, MSGNEW_HEIGHT
	SetColor 255, 255, 255
	DrawImage msgBg, 0, sy-MSGNEW_HEIGHT
	SetColor 0, 0, 0
	Local w# = TextWidth(inputContent), x# = (sx-w)*0.5
	DrawText inputContent, x, sy-(MSGNEW_HEIGHT-TextHeight(inputContent)/2)
	If blink > 25 Then

		DrawLine x+w+3, sy-MSGNEW_HEIGHT+10, x+w+3, sy-10

		If blink > 50 Then

			blink = 0

		EndIf

	EndIf
	blink :+ 1
		
	SetColor selectedSender.r, selectedSender.g, selectedSender.b
	DrawImage senderBg, 0, sy-MSGNEW_HEIGHT
	SetColor 255, 255, 255
	DrawText selectedSender.name, (MSG_PADDING-TextWidth(selectedSender.name))/2, sy-(MSGNEW_HEIGHT+TextHeight(selectedSender.name))/2

	If selectedKey Then

		SetColor keyR[selectedKey-1], keyG[selectedKey-1], keyB[selectedKey-1]

	Else

		SetColor 255, 255, 255

	EndIf
	DrawImage keyImg, MSG_PADDING, sy-(MSGNEW_HEIGHT+ImageHeight(keyImg))/2


	Local str$
	If selectedReceiver Then

		SetColor selectedReceiver.r, selectedReceiver.g, selectedReceiver.b
		str = selectedReceiver.name

	Else

		SetColor 100, 100, 100
		str = "*choose*"

	EndIf

	DrawImage receiverBg, sx-MSG_PADDING, sy-MSGNEW_HEIGHT
	SetColor 255, 255, 255
	DrawText str, sx-(MSG_PADDING+TextWidth(str))/2, sy-(MSGNEW_HEIGHT+TextHeight(str))/2

	If selectSender Or selectReceiver Then

		Local img:TImage, skip:TComputer
		If selectSender Then

			img = senderBg
			skip = selectedReceiver

		Else

			img = receiverBg
			skip = selectedSender

		EndIf

		SetColor 100, 100, 100
		DrawRect 0, sy-MSGNEW_HEIGHT-MSG_HEIGHT, sx, MSG_HEIGHT
		Local x#

		For Local c:TComputer = EachIn online

			If c = skip Then Continue

			SetColor c.r, c.g, c.b
			DrawImage img, x, sy-MSGNEW_HEIGHT-MSG_HEIGHT
			SetColor 255, 255, 255
			DrawText c.name, x+(MSG_PADDING-TextWidth(c.name))/2, sy-MSGNEW_HEIGHT-(MSG_HEIGHT+TextHeight(c.name))/2
			x :+ MSG_PADDING

		Next

	ElseIf selectKey Then

		Local x#
		SetColor 100, 100, 100
		DrawRect 0, sy-MSGNEW_HEIGHT-MSG_HEIGHT, sx, MSG_HEIGHT

		For Local i:Int = -1 Until NO_KEYS

			If i <> -1 Then
				If Not playerKeys[i] Then Continue
				SetColor keyR[i], keyG[i], keyB[i]
			Else
				SetColor 255, 255, 255
			EndIf
			DrawImage keyImg, x, sy-MSGNEW_HEIGHT-MSG_HEIGHT/2
			x :+ MSG_HEIGHT

		Next

	EndIf

End Function

Global permSelectSender:Int = False
Global permSend:Int = False
Global permCancel:Int = False
luaObject.Invoke "init", Null
TGameState.Load
OnEnd TGameState.Save

Local timer:TTimer = CreateTimer(60)
Repeat

	Select WaitEvent()

		Case EVENT_TIMERTICK

			WaitTimer timer
			If Not lockNotification Then UpdateEverything
			Cls
			DrawEverything
			Flip 0

		Case EVENT_APPTERMINATE

			End

		Case EVENT_KEYDOWN

			If EventData() = KEY_ESCAPE Then End

		Case EVENT_KEYCHAR

			If Not permSend Or lockNotification Then Continue

			Select EventData()

				Case 13
					If selectedSender And selectedReceiver Then
	
						Local msg:TMessage = CreateMessage(selectedSender, selectedReceiver, selectedKey, inputContent)

						inputContent = ""

					EndIf

				Case 8

					If inputContent Then

						inputContent = inputContent[..inputContent.length-1]

					EndIf

				Default

					If inputContent.length < MSG_MAX_LEN Then

						inputContent :+ Chr(EventData())

					EndIf

			End Select

		Case EVENT_MOUSEDOWN

			Select EventData()

				Case MOUSE_LEFT
				If Not permSend Or lockNotification Then Continue

				If (selectSender Or selectReceiver) And EventY() > sy-MSGNEW_HEIGHT-MSG_HEIGHT And EventY() < sy-MSGNEW_HEIGHT Then

					Local skip:TComputer
					If selectSender Then
						skip = selectedReceiver
					Else
						skip = selectedSender
					EndIf

					Local index:Int = EventX()/MSG_PADDING
					If index < online.Count() Then
						Local i:Int = 0
						For Local c:TComputer= EachIn online
							If c = skip Then Continue
							If i = index Then

								Select True
									Case selectSender
									selectedSender = c
									Case selectReceiver
									selectedReceiver = c
								End Select
							EndIf
							i :+ 1
						Next
					EndIf
					selectSender = False
					selectReceiver = False

				ElseIf selectKey And EventY() > sy-MSGNEW_HEIGHT-MSG_HEIGHT And EventY() < sy-MSGNEW_HEIGHT Then

					Local index:Int = EventX()/MSG_HEIGHT
					If index = 0 Then

						selectedKey = 0
						selectKey = False

					Else

						Local key:Int
						For key = 0 Until NO_KEYS
	
							If playerKeys[key] Then

								index :- 1
								If index = 0 Then Exit
	
							EndIf
	
						Next
						If index = 0 Then
	
							selectedKey = key+1
							selectKey = False
	
						EndIf

					EndIf

				ElseIf EventY() > sy-MSGNEW_HEIGHT Then

					If EventX() < MSG_PADDING Then

						selectReceiver = False
						selectSender = permSelectSender And Not selectSender

					ElseIf EventX() < MSG_PADDING+MSG_HEIGHT Then

						selectSender = False
						selectReceiver = False
						selectKey = True

					ElseIf EventX() > sx-MSG_PADDING Then

						selectSender = False
						selectReceiver = Not selectReceiver

					EndIf

				EndIf

				Case MOUSE_RIGHT

				If activeNotification Then

					activeNotification = Null
					lockNotification = False
					notiFadeStart = MilliSecs()
					Continue

				EndIf

				If selectSender Or selectReceiver Or selectKey Then

					selectSender = False
					selectReceiver = False
					selectKey = False
					Continue

				EndIf

				If Not permCancel Then Continue
				Local y# = 0
				For Local m:TMessage = EachIn messages

					If m.display Then
					
						y :+ MSG_HEIGHT

					Else

						y :+ m.space

					EndIf
					If y > EventY() Then
					
						m.Cancel
						Exit
						
					EndIf
					
				Next

			End Select
	End Select
Forever
